/*
**
** CNSsniff - version 2
** by cns at minithins.net
**
**
**
**
*/

/* preproc. */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <pcap.h>
#include <sys/param.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <dlfcn.h>

#include "cnssniff2.h"

#include <netinet/in_systm.h>
#ifndef __USE_BSD
#define __USE_BSD
#endif
#include <netinet/in.h>
#include <netinet/ip.h>

#define __FAVOR_BSD
#include <netinet/tcp.h>

/* global vars */
int force;
int mod; 
// 1 -| coute une interface
// 2 -| ouvre et lit un fichier
// 4 -| ouvre un dump et crit dedans
// 8 -| not taken
// 16 -| mod daemon
// 32 -| mod quiet
// 64 -| mod silent

int n_count;
char *dev;
char *dumpfile;
char *rule;
pcap_t *pcapdesc;
struct bpf_program fcode;
pcap_dumper_t *pcapdumper;
struct pcap_stat ps;
char errbuf[PCAP_ERRBUF_SIZE];
int count;
int snaplen, promisc, timeout;
time_t sigquit = 0;
int lhl;

// modules:
typedef struct modules { 
	struct modules *next;
	void *dlopened;
	void (*handle)(u_char *packet, int len, int lhl);
	} modules_t;

struct modules *heads = NULL;

static struct option long_options[] =
{
	{ "help",		0,	0,	'h'	},
	{ "interface", 	1, 	0, 	'i' },
	{ "read",		1,	0,	'r' },
	{ "write",		1,	0,	'w' },
	{ "promisc",	1,	0,	'p' },
	{ "quiet",		0,	0,	'q' },
	{ "count",		1,	0,	'c'	},
	{ "daemon",		0,	0,	'd'	},
	{ "force",		0,	0,	'f' },
	{ "snaplen",	1,	0,	's'	},
	{ "timeout",	1,	0,	't'	},
	{ "mod",		1,	0,	'm'	},
	{ "silent", 	0,	0,	'S' },
	{ "version",	0,	0,	'v'	},
	{ 0,			0,	0,	0	}
};

typedef struct datalink_type {
	char name[32];
	int lenght;
	} datalink_data;

const datalink_data datalink_main[] = {
	{ "no link-layer encapsulation", 4 },
	{ "Ethernet (10Mb)", 14 },
	{ "Experimental Ethernet (3Mb)", 14 },
	{ "Amateur Radio AX.25", 0 },            /* A REMPLIR */
	{ "Proteon ProNET Token Ring", 0 },      /* A REMPLIR */
	{ "Chaos", 0 },                          /* A REMPLIR */
	{ "IEEE 802 Networks", 0 },              /* A REMPLIR */
	{ "ARCNET", 0 },                         /* A REMPLIR */
	{ "Serial Line IP", 16 },
	{ "Point-to-point Protocol", 4 },
	{ "FDDI", 21 },
	{ "LLC/SNAP encapsulated atm", 0 },      /* A REMPLIR */
	{ "raw IP", 0 },
	{ "BSD/OS Serial Line IP", 16 },          /* should be ok... */
	{ "BSD/OS Point-to-point Protocol", 4 }, /* should be ok... */
	{ "napu", 0}
	};

/* functions */

void Version()
{
vlog(VERSION);
vlog(VERSION1);
vlog(VERSION2);
}

int load_module(char *mod)
{
struct modules *newm;
char chemin[2048];
char *error;

newm = (struct modules*)malloc(sizeof(struct modules));

snprintf(chemin, 2048, "%s/%s", getcwd(NULL, 0), mod);
printf("--[ %s\n", chemin);
if ((newm->dlopened = dlopen(chemin, RTLD_NOW)) == NULL)
	{
	error = dlerror();
	free(newm);
	evlog("dlopen error: %s\n", error);
	return 1;
	}

newm->handle = dlsym(newm->dlopened, "mod_fonction");
if ((error = dlerror()) != NULL)
	{
	evlog("dlsym: %s\n", error);
	free(newm);
	return 1;
	}

newm->next = heads;
heads = newm;

return(0);
}

char *datalink_name(int datalink_type)
{
char *nom;
vlog("type: %i\n", datalink_type);
nom = (char *)malloc(sizeof(char) * 32);
if (datalink_type >= 0 && datalink_type <= 15)
	{
	strcpy(nom, datalink_main[datalink_type].name);
	lhl = datalink_main[datalink_type].lenght;
	}
else
	{
	strcpy(nom, "not implemented");
	lhl = 0;
	}

return(nom);
}

void exiting()
{
if (mod & 4)
    pcap_dump_close(pcapdumper);
pcap_freecode(&fcode);
pcap_close(pcapdesc);
}

void gotsig(int sig)
{

time_t bla;
bla = time(NULL);

if ( bla < (sigquit + 2) || (mod & 16))
	{
	vlog("EXITING !!!\n");
	exiting();
	exit(0);
	}

show_stats();
sigquit = time(NULL);
}

void show_stats()
{
// requiere mod & 4 !
pcap_stats(pcapdesc,&ps);

vlog("Total: %6i recv, (%i processed) - %6i dropped\n",
	ps.ps_recv, n_count - 1, ps.ps_drop);

}

void showpacket(struct pcap_pkthdr hdr, const u_char *packet)
{
struct tm *mytime;
int i, j, k;

mytime = localtime(&hdr.ts.tv_sec);

vlog("%.2d:%.2d:%.2d.%.6i", mytime->tm_hour, mytime->tm_min,
	mytime->tm_sec, (int)hdr.ts.tv_usec);

vlog(", Len: %i (0x%x)\n",hdr.caplen,hdr.caplen);

for (i=0;i<=hdr.caplen/16;i++)
	{
	if(i*16 != hdr.caplen)
	vlog("0x%.2x  ",i*16);

	k =(((i+1)*16)>hdr.caplen)?hdr.caplen-(i*16):16;

	for (j=0;j<k;j++)
		vlog("%.2x ",*(packet++));

	if (k != 16)
		for (j=k;j<16;j++)
			vlog("   ");

	vlog("  ");
	packet-=k;

	for (j=0;j<k;j++)
		{
		if (*packet>20)
			vlog("%c",*packet);
		else
			vlog(".");
		packet++;
		}

	if (i*16 != hdr.caplen)
		vlog("\n");
	}
vlog("\n");
}

char *takerule(register char **argv)
{
register char **p;
register u_int len = 0;
char *buf;
char *src, *dst;
p = argv;
if (*p == 0)
	return 0;

while (*p)
	len += strlen(*p++) + 1;

if ((buf = (char *)malloc(len)) == NULL)
	{
	perror("malloc");
	exit(2);
	}

p = argv;
dst = buf;
while ((src = *p++) != NULL) {
	while ((*dst++ = *src++) != '\0');
	dst[-1] = ' ';
	}
dst[-1] = '\0';
return buf;
}


pcap_t *pcap_init_desc(int mod, char *arg)
{
pcap_t *temp;

if (mod & 1)
	{
	r_root();
	if ((temp = pcap_open_live (arg, snaplen, promisc, timeout, errbuf)) 
		== NULL)
		{
		evlog("pcap_open_live: %s\n", errbuf);
		exit(1);
		}
	return(temp);
	}

if (mod & 2)
	{
	if ((temp = pcap_open_offline (arg, errbuf)) == NULL)
		{
		evlog("pcap_open_offline: %s\n", errbuf);
		exit(1);
		} 
	return(temp);
	}

return(NULL);
}

void r_root()
{
if (getuid() && !force)
	{
	evlog("You're not root ... EXITING !!!\n");
	exit(1);
	}
}

int initvars(int argn, char **argv)
{
int arg, i;
int option_index = 0;

// initialisation de certaines variables.
dev = NULL; dumpfile = NULL;
count = -1;
mod = 0;
force = 0;
timeout = 500;
snaplen = 1024;
promisc = 1;

do
	{
	arg = getopt_long (argn, argv, "i:r:dw:p:qhc:fs:t:m:Sv",
		long_options, &option_index);

	switch (arg)
		{
		case 'v':
			Version();
			exit(0);
		break;
		case 'S':
			mod |= 64;	
		break;
		case 'm':
			load_module(optarg);
		break;
		case 's':
			snaplen = atoi(optarg);
		break;

		case 't':
			timeout = atoi(optarg);
		break;

		case 'f':
			force = 1;
		break;

		case 'q':
			mod |= 32;
		break;

		case 'c':
			count = atoi(optarg);
		break;

		case 'h':
			Usage();
			exit(0);
		break;

		case 'i': // choisir une interface.
			if (mod & 2)
				{
				evlog("Incompatible modes.\n");
				Usage();
				exit(1);
				}
			r_root();
			vlog("Changement de l'interface pour: %s\n", optarg); 
			dev = optarg;
			mod |= 1;
		break;
		
		case 'd':
			r_root();
			vlog("Toggling in daemon mode\n");
			if (fork () != 0)
				exit (0);
			if (setsid() == -1)
				perror("setsid");
			else
				{
				for (i = 0; i < NOFILE; i ++)
					if (i != 1 && i != 2)
						close(i);
				}
			mod |= 16;
		break;

		case 'r':
			if (mod & 1)
				{
				evlog("Incompatible modes.\n");
				exit(1);
				}
			mod |= 2;
			dev = optarg; 
			vlog("Reading file %s\n", dev);
		break;

		case 'w':
			dumpfile = optarg;
			vlog("Changing dump file to %s\n", dumpfile);
			mod |= 4;
		break;

		case 'p':
			promisc = atoi(optarg);
			vlog("promisc toggled to %i.\n", (promisc)?1:0);
		break;
		default:
		break;
		}
	}
while (arg != -1);

if (argv[optind] != NULL)
	{
	rule = takerule(&argv[optind]);
	vlog("Rule: [%s]\n", rule);
	}

if (!(mod & 1) && !(mod & 2))
	mod |= 1;

return 0;
}

int vlog(const char *sp, ...)
{
va_list vl;
if (mod & 32) 
	return 0;
va_start(vl, sp);
vprintf(sp, vl);
va_end(vl);
return(0);
}

int evlog(const char *sp, ...)
{
va_list vl;
FILE *stderr;
if (mod & 32)
	return 0;
stderr = fopen("/dev/stderr", "a+");
va_start(vl, sp);
vfprintf(stderr, sp, vl);
va_end(vl);
fclose (stderr);
return(0);
}

void Usage()
{
vlog(VERSION);
vlog(VERSION1);
vlog(VERSION2);
vlog("./cnssniff2 <options> [expression BPF]\n");
vlog("-i\t--interface\t\tChoix de l'interface\n");
vlog("-c\t--count    \t\tFiltrer n paquets\n");
vlog("-r\t--read     \t\tLire un fichier de dump\n");
vlog("-w\t--write    \t\tEcrire dans un fichier de dump\n");
vlog("-q\t--quiet    \t\tMode silencieux\n");
vlog("-d\t--daemon   \t\tMode daemon\n");
vlog("-t\t--timeout  \t\tChanger la valeur de timeout\n");
vlog("-s\t--snaplen  \t\tTaille de la capture des paquets\n");
vlog("-p\t--promisc  \t\tPromiscious mode ?\n");
vlog("-f\t--force    \t\tForcer la vrification uid = 0\n"); 
vlog("-h\t--help     \t\tEcran d'aide\n");
vlog("-m\t--mod      \t\tCharger un module\n");
vlog("-v\t--version  \t\tAfficher la version\n");
vlog("\n");
}

int main (int argn, char **argv, char **envp)
{
struct modules *tmp;
u_char *my_packet;
struct pcap_pkthdr usefull;
bpf_u_int32 localnet, netmask;
struct sigaction myend;

myend.sa_handler = (void*)gotsig;
myend.sa_flags = 0 ;
sigaction(SIGINT,&myend,NULL);

initvars(argn, argv);

if (dev == NULL && (mod & 1))
	{
	if ((dev = pcap_lookupdev(errbuf)) == NULL)
		{
		evlog("pcap_lookupdev: %s\n", errbuf);
		exit (1);
		}
	}

pcapdesc = pcap_init_desc(mod, dev);

if ((mod & 4) && dumpfile)
	if ((pcapdumper = pcap_dump_open(pcapdesc, dumpfile)) == NULL)
		evlog("pcap_dump_open: %s\n", pcap_geterr(pcapdesc));

if (mod & 1)
	{
	if (pcap_lookupnet(dev, &localnet, &netmask, errbuf)<0)
		evlog("pcap_lookupnet: %s\n", errbuf);
	}
else
	localnet = 0x00000000;

if (pcap_compile(pcapdesc,&fcode,rule,1,localnet)<0)
	evlog("pcap_compile: %s\n", pcap_geterr(pcapdesc));
if (pcap_setfilter(pcapdesc, &fcode) < 0)
	evlog("pcap_setfilter: %s\n", pcap_geterr(pcapdesc));

vlog("Datalink Type: %s\n", datalink_name(pcap_datalink(pcapdesc)));

if (mod & 1)
	{
	if (pcap_stats(pcapdesc, &ps))
		evlog("pcap_stats: %s\n", pcap_geterr(pcapdesc));
	}

n_count = 1;
do {
	my_packet = (u_char *) pcap_next(pcapdesc, &usefull);
	
	if (my_packet != NULL)
		{ 

		if (mod & 4)
			pcap_dump((u_char*)pcapdumper, &usefull, my_packet);
		if (!(mod & 64))
			showpacket(usefull, my_packet);
		//evlog("CNS2: dbut modules\n");
		// new module support
		for (tmp = heads; tmp != NULL; tmp = tmp->next)
			tmp->handle(my_packet, usefull.len, lhl);
		//evlog("CNS2: fin modules\n");

		n_count ++;
		}

} while ( ( (mod & 1) && ((n_count <= count)||count == -1) )
			|| ( (mod & 2) && ((n_count <= count)||count == -1)
			&& (my_packet != NULL) ));

show_stats();

exiting();

return 0;
}
